@@ -1,31 +1,39 @@
-require 'date'
-require 'set'
-require 'bigdecimal'
-require 'bigdecimal/util'
-
 module ActiveRecord
   module ConnectionAdapters #:nodoc:
     # Abstract representation of an index definition on a table. Instances of
     # this type are typically created and returned by methods in database
     # adapters. e.g. ActiveRecord::ConnectionAdapters::AbstractMysqlAdapter#indexes
-    class IndexDefinition < Struct.new(:table, :name, :unique, :columns, :lengths, :orders, :where, :type, :using) #:nodoc:
-    end
+    IndexDefinition = Struct.new(:table, :name, :unique, :columns, :lengths, :orders, :where, :type, :using, :comment) #:nodoc:
 
     # Abstract representation of a column definition. Instances of this type
     # are typically created by methods in TableDefinition, and added to the
     # +columns+ attribute of said TableDefinition object, in order to be used
     # for generating a number of table creation or table changing SQL statements.
-    class ColumnDefinition < Struct.new(:name, :type, :limit, :precision, :scale, :default, :null, :first, :after, :primary_key, :sql_type, :cast_type) #:nodoc:
-
+    ColumnDefinition = Struct.new(:name, :type, :options, :sql_type) do # :nodoc:
       def primary_key?
-        primary_key || type.to_sym == :primary_key
+        options[:primary_key]
+      end
+
+      [:limit, :precision, :scale, :default, :null, :collation, :comment].each do |option_name|
+        module_eval <<-CODE, __FILE__, __LINE__ + 1
+          def #{option_name}
+            options[:#{option_name}]
+          end
+
+          def #{option_name}=(value)
+            options[:#{option_name}] = value
+          end
+        CODE
       end
     end
 
-    class ChangeColumnDefinition < Struct.new(:column, :type, :options) #:nodoc:
-    end
+    AddColumnDefinition = Struct.new(:column) # :nodoc:
 
-    class ForeignKeyDefinition < Struct.new(:from_table, :to_table, :options) #:nodoc:
+    ChangeColumnDefinition = Struct.new(:column, :name) #:nodoc:
+
+    PrimaryKeyDefinition = Struct.new(:name) # :nodoc:
+
+    ForeignKeyDefinition = Struct.new(:from_table, :to_table, :options) do #:nodoc:
       def name
         options[:name]
       end
@@ -50,31 +58,147 @@ module ActiveRecord
         options[:primary_key] != default_primary_key
       end
 
-      private
-      def default_primary_key
-        "id"
+      def defined_for?(to_table_ord = nil, to_table: nil, **options)
+        if to_table_ord
+          self.to_table == to_table_ord.to_s
+        else
+          (to_table.nil? || to_table.to_s == self.to_table) &&
+            options.all? { |k, v| self.options[k].to_s == v.to_s }
+        end
       end
+
+      private
+        def default_primary_key
+          "id"
+        end
     end
 
-    module TimestampDefaultDeprecation # :nodoc:
-      def emit_warning_if_null_unspecified(sym, options)
-        return if options.key?(:null)
+    class ReferenceDefinition # :nodoc:
+      def initialize(
+        name,
+        polymorphic: false,
+        index: true,
+        foreign_key: false,
+        type: :bigint,
+        **options
+      )
+        @name = name
+        @polymorphic = polymorphic
+        @index = index
+        @foreign_key = foreign_key
+        @type = type
+        @options = options
 
-        ActiveSupport::Deprecation.warn(<<-MSG.squish)
-          `##{sym}` was called without specifying an option for `null`. In Rails 5,
-          this behavior will change to `null: false`. You should manually specify
-         `null: true` to prevent the behavior of your existing migrations from changing.
-        MSG
+        if polymorphic && foreign_key
+          raise ArgumentError, "Cannot add a foreign key to a polymorphic relation"
+        end
       end
+
+      def add_to(table)
+        columns.each do |column_options|
+          table.column(*column_options)
+        end
+
+        if index
+          table.index(column_names, index_options)
+        end
+
+        if foreign_key
+          table.foreign_key(foreign_table_name, foreign_key_options)
+        end
+      end
+
+      # TODO Change this to private once we've dropped Ruby 2.2 support.
+      # Workaround for Ruby 2.2 "private attribute?" warning.
+      protected
+
+        attr_reader :name, :polymorphic, :index, :foreign_key, :type, :options
+
+      private
+
+        def as_options(value)
+          value.is_a?(Hash) ? value : {}
+        end
+
+        def polymorphic_options
+          as_options(polymorphic).merge(null: options[:null])
+        end
+
+        def index_options
+          as_options(index)
+        end
+
+        def foreign_key_options
+          as_options(foreign_key).merge(column: column_name)
+        end
+
+        def columns
+          result = [[column_name, type, options]]
+          if polymorphic
+            result.unshift(["#{name}_type", :string, polymorphic_options])
+          end
+          result
+        end
+
+        def column_name
+          "#{name}_id"
+        end
+
+        def column_names
+          columns.map(&:first)
+        end
+
+        def foreign_table_name
+          foreign_key_options.fetch(:to_table) do
+            Base.pluralize_table_names ? name.to_s.pluralize : name
+          end
+        end
+    end
+
+    module ColumnMethods
+      # Appends a primary key definition to the table definition.
+      # Can be called multiple times, but this is probably not a good idea.
+      def primary_key(name, type = :primary_key, **options)
+        column(name, type, options.merge(primary_key: true))
+      end
+
+      # Appends a column or columns of a specified type.
+      #
+      #  t.string(:goat)
+      #  t.string(:goat, :sheep)
+      #
+      # See TableDefinition#column
+      [
+        :bigint,
+        :binary,
+        :boolean,
+        :date,
+        :datetime,
+        :decimal,
+        :float,
+        :integer,
+        :string,
+        :text,
+        :time,
+        :timestamp,
+        :virtual,
+      ].each do |column_type|
+        module_eval <<-CODE, __FILE__, __LINE__ + 1
+          def #{column_type}(*args, **options)
+            args.each { |name| column(name, :#{column_type}, options) }
+          end
+        CODE
+      end
+      alias_method :numeric, :decimal
     end
 
     # Represents the schema of an SQL table in an abstract way. This class
     # provides methods for manipulating the schema representation.
     #
-    # Inside migration files, the +t+ object in +create_table+
+    # Inside migration files, the +t+ object in {create_table}[rdoc-ref:SchemaStatements#create_table]
     # is actually of this type:
     #
-    #   class SomeMigration < ActiveRecord::Migration
+    #   class SomeMigration < ActiveRecord::Migration[5.0]
     #     def up
     #       create_table :foo do |t|
     #         puts t.class  # => "ActiveRecord::ConnectionAdapters::TableDefinition"
@@ -86,125 +210,55 @@ module ActiveRecord
     #     end
     #   end
     #
-    # The table definitions
-    # The Columns are stored as a ColumnDefinition in the +columns+ attribute.
     class TableDefinition
-      include TimestampDefaultDeprecation
+      include ColumnMethods
 
-      # An array of ColumnDefinition objects, representing the column changes
-      # that have been defined.
       attr_accessor :indexes
-      attr_reader :name, :temporary, :options, :as, :foreign_keys
+      attr_reader :name, :temporary, :options, :as, :foreign_keys, :comment
 
-      def initialize(types, name, temporary, options, as = nil)
+      def initialize(name, temporary = false, options = nil, as = nil, comment: nil)
         @columns_hash = {}
-        @indexes = {}
-        @foreign_keys = {}
-        @native = types
+        @indexes = []
+        @foreign_keys = []
+        @primary_keys = nil
         @temporary = temporary
         @options = options
         @as = as
         @name = name
+        @comment = comment
       end
 
+      def primary_keys(name = nil) # :nodoc:
+        @primary_keys = PrimaryKeyDefinition.new(name) if name
+        @primary_keys
+      end
+
+      # Returns an array of ColumnDefinition objects for the columns of the table.
       def columns; @columns_hash.values; end
 
-      # Appends a primary key definition to the table definition.
-      # Can be called multiple times, but this is probably not a good idea.
-      def primary_key(name, type = :primary_key, options = {})
-        column(name, type, options.merge(:primary_key => true))
-      end
-
       # Returns a ColumnDefinition for the column with name +name+.
       def [](name)
         @columns_hash[name.to_s]
       end
 
       # Instantiates a new column for the table.
-      # The +type+ parameter is normally one of the migrations native types,
-      # which is one of the following:
-      # <tt>:primary_key</tt>, <tt>:string</tt>, <tt>:text</tt>,
-      # <tt>:integer</tt>, <tt>:float</tt>, <tt>:decimal</tt>,
-      # <tt>:datetime</tt>, <tt>:time</tt>, <tt>:date</tt>,
-      # <tt>:binary</tt>, <tt>:boolean</tt>.
+      # See {connection.add_column}[rdoc-ref:ConnectionAdapters::SchemaStatements#add_column]
+      # for available options.
       #
-      # You may use a type not in this list as long as it is supported by your
-      # database (for example, "polygon" in MySQL), but this will not be database
-      # agnostic and should usually be avoided.
-      #
-      # Available options are (none of these exists by default):
-      # * <tt>:limit</tt> -
-      #   Requests a maximum column length. This is number of characters for <tt>:string</tt> and
-      #   <tt>:text</tt> columns and number of bytes for <tt>:binary</tt> and <tt>:integer</tt> columns.
-      # * <tt>:default</tt> -
-      #   The column's default value. Use nil for NULL.
-      # * <tt>:null</tt> -
-      #   Allows or disallows +NULL+ values in the column. This option could
-      #   have been named <tt>:null_allowed</tt>.
-      # * <tt>:precision</tt> -
-      #   Specifies the precision for a <tt>:decimal</tt> column.
-      # * <tt>:scale</tt> -
-      #   Specifies the scale for a <tt>:decimal</tt> column.
+      # Additional options are:
       # * <tt>:index</tt> -
       #   Create an index for the column. Can be either <tt>true</tt> or an options hash.
       #
-      # Note: The precision is the total number of significant digits
-      # and the scale is the number of digits that can be stored following
-      # the decimal point. For example, the number 123.45 has a precision of 5
-      # and a scale of 2. A decimal with a precision of 5 and a scale of 2 can
-      # range from -999.99 to 999.99.
-      #
-      # Please be aware of different RDBMS implementations behavior with
-      # <tt>:decimal</tt> columns:
-      # * The SQL standard says the default scale should be 0, <tt>:scale</tt> <=
-      #   <tt>:precision</tt>, and makes no comments about the requirements of
-      #   <tt>:precision</tt>.
-      # * MySQL: <tt>:precision</tt> [1..63], <tt>:scale</tt> [0..30].
-      #   Default is (10,0).
-      # * PostgreSQL: <tt>:precision</tt> [1..infinity],
-      #   <tt>:scale</tt> [0..infinity]. No default.
-      # * SQLite2: Any <tt>:precision</tt> and <tt>:scale</tt> may be used.
-      #   Internal storage as strings. No default.
-      # * SQLite3: No restrictions on <tt>:precision</tt> and <tt>:scale</tt>,
-      #   but the maximum supported <tt>:precision</tt> is 16. No default.
-      # * Oracle: <tt>:precision</tt> [1..38], <tt>:scale</tt> [-84..127].
-      #   Default is (38,0).
-      # * DB2: <tt>:precision</tt> [1..63], <tt>:scale</tt> [0..62].
-      #   Default unknown.
-      # * SqlServer?: <tt>:precision</tt> [1..38], <tt>:scale</tt> [0..38].
-      #   Default (38,0).
-      #
       # This method returns <tt>self</tt>.
       #
       # == Examples
+      #
       #  # Assuming +td+ is an instance of TableDefinition
-      #  td.column(:granted, :boolean)
-      #  # granted BOOLEAN
-      #
-      #  td.column(:picture, :binary, limit: 2.megabytes)
-      #  # => picture BLOB(2097152)
-      #
-      #  td.column(:sales_stage, :string, limit: 20, default: 'new', null: false)
-      #  # => sales_stage VARCHAR(20) DEFAULT 'new' NOT NULL
-      #
-      #  td.column(:bill_gates_money, :decimal, precision: 15, scale: 2)
-      #  # => bill_gates_money DECIMAL(15,2)
-      #
-      #  td.column(:sensor_reading, :decimal, precision: 30, scale: 20)
-      #  # => sensor_reading DECIMAL(30,20)
-      #
-      #  # While <tt>:scale</tt> defaults to zero on most databases, it
-      #  # probably wouldn't hurt to include it.
-      #  td.column(:huge_integer, :decimal, precision: 30)
-      #  # => huge_integer DECIMAL(30)
-      #
-      #  # Defines a column with a database-specific type.
-      #  td.column(:foo, 'polygon')
-      #  # => foo polygon
+      #  td.column(:granted, :boolean, index: true)
       #
       # == Short-hand examples
       #
-      # Instead of calling +column+ directly, you can also work with the short-hand definitions for the default types.
+      # Instead of calling #column directly, you can also work with the short-hand definitions for the default types.
       # They use the type as the method name instead of as a parameter and allow for multiple columns to be defined
       # in a single statement.
       #
@@ -236,7 +290,8 @@ module ActiveRecord
       # TableDefinition#references will add an appropriately-named _id column, plus a corresponding _type
       # column if the <tt>:polymorphic</tt> option is supplied. If <tt>:polymorphic</tt> is a hash of
       # options, these will be used when creating the <tt>_type</tt> column. The <tt>:index</tt> option
-      # will also create an index, similar to calling <tt>add_index</tt>. So what can be written like this:
+      # will also create an index, similar to calling {add_index}[rdoc-ref:ConnectionAdapters::SchemaStatements#add_index].
+      # So what can be written like this:
       #
       #   create_table :taggings do |t|
       #     t.integer :tag_id, :tagger_id, :taggable_id
@@ -255,7 +310,7 @@ module ActiveRecord
       #   end
       def column(name, type, options = {})
         name = name.to_s
-        type = type.to_sym
+        type = type.to_sym if type
         options = options.dup
 
         if @columns_hash[name] && @columns_hash[name].primary_key?
@@ -268,37 +323,34 @@ module ActiveRecord
         self
       end
 
+      # remove the column +name+ from the table.
+      #   remove_column(:account_id)
       def remove_column(name)
         @columns_hash.delete name.to_s
       end
 
-      [:string, :text, :integer, :bigint, :float, :decimal, :datetime, :timestamp, :time, :date, :binary, :boolean].each do |column_type|
-        define_method column_type do |*args|
-          options = args.extract_options!
-          column_names = args
-          column_names.each { |name| column(name, column_type, options) }
-        end
-      end
-
       # Adds index options to the indexes hash, keyed by column name
       # This is primarily used to track indexes that need to be created after the table
       #
       #   index(:account_id, name: 'index_projects_on_account_id')
       def index(column_name, options = {})
-        indexes[column_name] = options
+        indexes << [column_name, options]
       end
 
       def foreign_key(table_name, options = {}) # :nodoc:
-        foreign_keys[table_name] = options
+        table_name_prefix = ActiveRecord::Base.table_name_prefix
+        table_name_suffix = ActiveRecord::Base.table_name_suffix
+        table_name = "#{table_name_prefix}#{table_name}#{table_name_suffix}"
+        foreign_keys.push([table_name, options])
       end
 
       # Appends <tt>:datetime</tt> columns <tt>:created_at</tt> and
-      # <tt>:updated_at</tt> to the table. See SchemaStatements#add_timestamps
+      # <tt>:updated_at</tt> to the table. See {connection.add_timestamps}[rdoc-ref:SchemaStatements#add_timestamps]
       #
       #   t.timestamps null: false
-      def timestamps(*args)
-        options = args.extract_options!
-        emit_warning_if_null_unspecified(:timestamps, options)
+      def timestamps(**options)
+        options[:null] = false if options[:null].nil?
+
         column(:created_at, :datetime, options)
         column(:updated_at, :datetime, options)
       end
@@ -308,60 +360,29 @@ module ActiveRecord
       #  t.references(:user)
       #  t.belongs_to(:supplier, foreign_key: true)
       #
-      # See SchemaStatements#add_reference for details of the options you can use.
-      def references(*args)
-        options = args.extract_options!
-        polymorphic = options.delete(:polymorphic)
-        index_options = options.delete(:index)
-        foreign_key_options = options.delete(:foreign_key)
-        type = options.delete(:type) || :integer
-
-        if polymorphic && foreign_key_options
-          raise ArgumentError, "Cannot add a foreign key on a polymorphic relation"
-        end
-
-        args.each do |col|
-          column("#{col}_id", type, options)
-          column("#{col}_type", :string, polymorphic.is_a?(Hash) ? polymorphic : options) if polymorphic
-          index(polymorphic ? %w(type id).map { |t| "#{col}_#{t}" } : "#{col}_id", index_options.is_a?(Hash) ? index_options : {}) if index_options
-          if foreign_key_options
-            to_table = Base.pluralize_table_names ? col.to_s.pluralize : col.to_s
-            foreign_key(to_table, foreign_key_options.is_a?(Hash) ? foreign_key_options : {})
-          end
+      # See {connection.add_reference}[rdoc-ref:SchemaStatements#add_reference] for details of the options you can use.
+      def references(*args, **options)
+        args.each do |ref_name|
+          ReferenceDefinition.new(ref_name, options).add_to(self)
         end
       end
       alias :belongs_to :references
 
-      def new_column_definition(name, type, options) # :nodoc:
+      def new_column_definition(name, type, **options) # :nodoc:
         type = aliased_types(type.to_s, type)
-        column = create_column_definition name, type
-        limit = options.fetch(:limit) do
-          native[type][:limit] if native[type].is_a?(Hash)
-        end
-
-        column.limit       = limit
-        column.precision   = options[:precision]
-        column.scale       = options[:scale]
-        column.default     = options[:default]
-        column.null        = options[:null]
-        column.first       = options[:first]
-        column.after       = options[:after]
-        column.primary_key = type == :primary_key || options[:primary_key]
-        column
+        options[:primary_key] ||= type == :primary_key
+        options[:null] = false if options[:primary_key]
+        create_column_definition(name, type, options)
       end
 
       private
-      def create_column_definition(name, type)
-        ColumnDefinition.new name, type
-      end
+        def create_column_definition(name, type, options)
+          ColumnDefinition.new(name, type, options)
+        end
 
-      def native
-        @native
-      end
-
-      def aliased_types(name, fallback)
-        'timestamp' == name ? :datetime : fallback
-      end
+        def aliased_types(name, fallback)
+          "timestamp" == name ? :datetime : fallback
+        end
     end
 
     class AlterTable # :nodoc:
@@ -389,16 +410,17 @@ module ActiveRecord
       def add_column(name, type, options)
         name = name.to_s
         type = type.to_sym
-        @adds << @td.new_column_definition(name, type, options)
+        @adds << AddColumnDefinition.new(@td.new_column_definition(name, type, options))
       end
     end
 
     # Represents an SQL table in an abstract way for updating a table.
-    # Also see TableDefinition and SchemaStatements#create_table
+    # Also see TableDefinition and {connection.create_table}[rdoc-ref:SchemaStatements#create_table]
     #
     # Available transformations are:
     #
     #   change_table :table do |t|
+    #     t.primary_key
     #     t.column
     #     t.index
     #     t.rename_index
@@ -411,8 +433,10 @@ module ActiveRecord
     #     t.string
     #     t.text
     #     t.integer
+    #     t.bigint
     #     t.float
     #     t.decimal
+    #     t.numeric
     #     t.datetime
     #     t.timestamp
     #     t.time
@@ -427,6 +451,8 @@ module ActiveRecord
     #   end
     #
     class Table
+      include ColumnMethods
+
       attr_reader :name
 
       def initialize(table_name, base)
@@ -435,33 +461,42 @@ module ActiveRecord
       end
 
       # Adds a new column to the named table.
-      # See TableDefinition#column for details of the options you can use.
       #
-      # ====== Creating a simple column
       #  t.column(:name, :string)
+      #
+      # See TableDefinition#column for details of the options you can use.
       def column(column_name, type, options = {})
         @base.add_column(name, column_name, type, options)
       end
 
-      # Checks to see if a column exists. See SchemaStatements#column_exists?
+      # Checks to see if a column exists.
+      #
+      #  t.string(:name) unless t.column_exists?(:name, :string)
+      #
+      # See {connection.column_exists?}[rdoc-ref:SchemaStatements#column_exists?]
       def column_exists?(column_name, type = nil, options = {})
         @base.column_exists?(name, column_name, type, options)
       end
 
       # Adds a new index to the table. +column_name+ can be a single Symbol, or
-      # an Array of Symbols. See SchemaStatements#add_index
+      # an Array of Symbols.
       #
-      # ====== Creating a simple index
       #  t.index(:name)
-      # ====== Creating a unique index
       #  t.index([:branch_id, :party_id], unique: true)
-      # ====== Creating a named index
       #  t.index([:branch_id, :party_id], unique: true, name: 'by_branch_party')
+      #
+      # See {connection.add_index}[rdoc-ref:SchemaStatements#add_index] for details of the options you can use.
       def index(column_name, options = {})
         @base.add_index(name, column_name, options)
       end
 
-      # Checks to see if an index exists. See SchemaStatements#index_exists?
+      # Checks to see if an index exists.
+      #
+      #  unless t.index_exists?(:branch_id)
+      #    t.index(:branch_id)
+      #  end
+      #
+      # See {connection.index_exists?}[rdoc-ref:SchemaStatements#index_exists?]
       def index_exists?(column_name, options = {})
         @base.index_exists?(name, column_name, options)
       end
@@ -469,52 +504,59 @@ module ActiveRecord
       # Renames the given index on the table.
       #
       #  t.rename_index(:user_id, :account_id)
+      #
+      # See {connection.rename_index}[rdoc-ref:SchemaStatements#rename_index]
       def rename_index(index_name, new_index_name)
         @base.rename_index(name, index_name, new_index_name)
       end
 
-      # Adds timestamps (+created_at+ and +updated_at+) columns to the table. See SchemaStatements#add_timestamps
+      # Adds timestamps (+created_at+ and +updated_at+) columns to the table.
       #
-      #  t.timestamps null: false
+      #  t.timestamps(null: false)
+      #
+      # See {connection.add_timestamps}[rdoc-ref:SchemaStatements#add_timestamps]
       def timestamps(options = {})
         @base.add_timestamps(name, options)
       end
 
       # Changes the column's definition according to the new options.
-      # See TableDefinition#column for details of the options you can use.
       #
       #  t.change(:name, :string, limit: 80)
       #  t.change(:description, :text)
+      #
+      # See TableDefinition#column for details of the options you can use.
       def change(column_name, type, options = {})
         @base.change_column(name, column_name, type, options)
       end
 
-      # Sets a new default value for a column. See SchemaStatements#change_column_default
+      # Sets a new default value for a column.
       #
       #  t.change_default(:qualification, 'new')
       #  t.change_default(:authorized, 1)
-      def change_default(column_name, default)
-        @base.change_column_default(name, column_name, default)
+      #  t.change_default(:status, from: nil, to: "draft")
+      #
+      # See {connection.change_column_default}[rdoc-ref:SchemaStatements#change_column_default]
+      def change_default(column_name, default_or_changes)
+        @base.change_column_default(name, column_name, default_or_changes)
       end
 
       # Removes the column(s) from the table definition.
       #
       #  t.remove(:qualification)
       #  t.remove(:qualification, :experience)
+      #
+      # See {connection.remove_columns}[rdoc-ref:SchemaStatements#remove_columns]
       def remove(*column_names)
         @base.remove_columns(name, *column_names)
       end
 
       # Removes the given index from the table.
       #
-      # ====== Remove the index_table_name_on_column in the table_name table
-      #   t.remove_index :column
-      # ====== Remove the index named index_table_name_on_branch_id in the table_name table
-      #   t.remove_index column: :branch_id
-      # ====== Remove the index named index_table_name_on_branch_id_and_party_id in the table_name table
-      #   t.remove_index column: [:branch_id, :party_id]
-      # ====== Remove the index named by_branch_party in the table_name table
-      #   t.remove_index name: :by_branch_party
+      #   t.remove_index(:branch_id)
+      #   t.remove_index(column: [:branch_id, :party_id])
+      #   t.remove_index(name: :by_branch_party)
+      #
+      # See {connection.remove_index}[rdoc-ref:SchemaStatements#remove_index]
       def remove_index(options = {})
         @base.remove_index(name, options)
       end
@@ -522,6 +564,8 @@ module ActiveRecord
       # Removes the timestamp columns (+created_at+ and +updated_at+) from the table.
       #
       #  t.remove_timestamps
+      #
+      # See {connection.remove_timestamps}[rdoc-ref:SchemaStatements#remove_timestamps]
       def remove_timestamps(options = {})
         @base.remove_timestamps(name, options)
       end
@@ -529,6 +573,8 @@ module ActiveRecord
       # Renames a column.
       #
       #  t.rename(:description, :name)
+      #
+      # See {connection.rename_column}[rdoc-ref:SchemaStatements#rename_column]
       def rename(column_name, new_column_name)
         @base.rename_column(name, column_name, new_column_name)
       end
@@ -538,9 +584,8 @@ module ActiveRecord
       #  t.references(:user)
       #  t.belongs_to(:supplier, foreign_key: true)
       #
-      # See SchemaStatements#add_reference for details of the options you can use.
-      def references(*args)
-        options = args.extract_options!
+      # See {connection.add_reference}[rdoc-ref:SchemaStatements#add_reference] for details of the options you can use.
+      def references(*args, **options)
         args.each do |ref_name|
           @base.add_reference(name, ref_name, options)
         end
@@ -548,37 +593,35 @@ module ActiveRecord
       alias :belongs_to :references
 
       # Removes a reference. Optionally removes a +type+ column.
-      # <tt>remove_references</tt> and <tt>remove_belongs_to</tt> are acceptable.
       #
       #  t.remove_references(:user)
       #  t.remove_belongs_to(:supplier, polymorphic: true)
       #
-      # See SchemaStatements#remove_reference
-      def remove_references(*args)
-        options = args.extract_options!
+      # See {connection.remove_reference}[rdoc-ref:SchemaStatements#remove_reference]
+      def remove_references(*args, **options)
         args.each do |ref_name|
           @base.remove_reference(name, ref_name, options)
         end
       end
       alias :remove_belongs_to :remove_references
 
-      # Adds a column or columns of a specified type
+      # Adds a foreign key.
       #
-      #  t.string(:goat)
-      #  t.string(:goat, :sheep)
-      [:string, :text, :integer, :float, :decimal, :datetime, :timestamp, :time, :date, :binary, :boolean].each do |column_type|
-        define_method column_type do |*args|
-          options = args.extract_options!
-          args.each do |column_name|
-            @base.add_column(name, column_name, column_type, options)
-          end
-        end
+      # t.foreign_key(:authors)
+      #
+      # See {connection.add_foreign_key}[rdoc-ref:SchemaStatements#add_foreign_key]
+      def foreign_key(*args) # :nodoc:
+        @base.add_foreign_key(name, *args)
       end
 
-      private
-        def native
-          @base.native_database_types
-        end
+      # Checks to see if a foreign key exists.
+      #
+      # t.foreign_key(:authors) unless t.foreign_key_exists?(:authors)
+      #
+      # See {connection.foreign_key_exists?}[rdoc-ref:SchemaStatements#foreign_key_exists?]
+      def foreign_key_exists?(*args) # :nodoc:
+        @base.foreign_key_exists?(name, *args)
+      end
     end
   end
 end
