@@ -5,10 +5,10 @@ module ActiveRecord
   end
 
   module VERSION
-    MAJOR = 4
-    MINOR = 2
-    TINY  = 5
-    PRE   = "2"
+    MAJOR = 5
+    MINOR = 1
+    TINY  = 4
+    PRE   = nil
 
     STRING = [MAJOR, MINOR, TINY, PRE].compact.join(".")
   end
